-- Create an APL trait for the APL class
---@class APLTrait
---@field cb fun(actor?: APLActorTable):boolean
---@field lastcall number
local APLTrait = {}
APLTrait.__index = APLTrait

-- Constructor
---@param cb fun(actor?: APLActorTable):boolean
---@return APLTrait
function APLTrait:New(cb)
    local self = setmetatable({}, APLTrait)

    self.cb = cb
    self.lastcall = 0

    return self
end

-- Evaulate the APL trait
---@param actor? APLActorTable
---@return boolean
function APLTrait:Evaluate(actor)
    if GetTime() - self.lastcall > 0.1 then
        self.lastresult = self.cb(actor)
        self.lastcall = GetTime()
        return self.lastresult
    end

    return self.lastresult
end

-- tostring
---@return string
function APLTrait:__tostring()
    return "Bastion.__APLTrait"
end

---@class APLActorTableBase
---@field type "spell" | "item" | "apl" | "sequencer" | "variable" | "action"

---@alias APLActorTable APLActorSpellTable | APLActorItemTable | APLActorAPLTable | APLActorSequencerTable | APLActorVariableTable | APLActorActionTable

-- Create an APL actor for the APL class
---@class APLActor
---@field actor APLActorTable
---@field traits APLTrait[]
---@field name string
local APLActor = {}
APLActor.__index = APLActor

-- Constructor
---@param actor APLActorTable
function APLActor:New(actor)
    local self = setmetatable({}, APLActor)
    if actor.type == "spell" then
        self.name = string.format("[%s] `%s`<%s>", "spell", actor.spell:GetName(), actor.spell:GetID())
    elseif actor.type == "item" then
        self.name = string.format("[%s] `%s`<%s>", "item", actor.item:GetName(), actor.item:GetID())
    elseif actor.type == "apl" then
        self.name = string.format("[%s] `%s`", "apl", actor.apl.name)
    elseif actor.type == "sequencer" then
        self.name = string.format("[%s]", "sequencer")
    elseif actor.type == "variable" then
        self.name = string.format("[%s] `%s`", "variable", actor.variable)
    elseif actor.type == "action" then
        self.name = string.format("[%s] `%s`", "action", actor.action)
    else
        self.name = string.format("[%s] Unknown", actor.type)
    end
    self.actor = actor
    self.traits = {}

    return self
end

-- Add a trait to the APL actor
---@param ... APLTrait
---@return APLActor
function APLActor:AddTraits(...)
    for _, trait in ipairs({ ... }) do
        table.insert(self.traits, trait)
    end

    return self
end

-- Get the actor
---@return APLActorTable
function APLActor:GetActor()
    return self.actor
end

-- Evaulate the APL actor
---@return boolean
function APLActor:Evaluate()
    for _, trait in ipairs(self.traits) do
        if not trait:Evaluate(self:GetActor()) then
            return false
        end
    end

    return true
end

-- Execute
function APLActor:Execute()
    local actorTable = self:GetActor()
    -- If the actor is a sequencer we don't want to continue executing the APL if the sequencer is not finished
    if actorTable.type == "sequencer" then
        ---@cast actorTable APLActorSequencerTable
        if actorTable.condition and actorTable.condition() and not actorTable.sequencer:Finished() then
            actorTable.sequencer:Execute()
            return true
        end

        if not actorTable.condition and not actorTable.sequencer:Finished() then
            actorTable.sequencer:Execute()
            return true
        end

        -- Check if the sequencer can be reset and reset it if it can
        if actorTable.sequencer:ShouldReset() then
            actorTable.sequencer:Reset()
        end
    end
    if actorTable.type == "apl" then
        ---@cast actorTable APLActorAPLTable
        if actorTable.condition and actorTable.condition() then
            -- print("Bastion: APL:Execute: Executing sub APL " .. actorTable.apl.name)
            if actorTable.apl:Execute() then
                return true
            end
        end
    end
    if actorTable.type == "spell" then
        ---@cast actorTable APLActorSpellTable
        return actorTable.spell:CastableIf(actorTable.castableFunc):OnCast(actorTable.onCastFunc):Cast(actorTable.target, actorTable.condition)
    end
    if actorTable.type == "item" then
        ---@cast actorTable APLActorItemTable
        return actorTable.item:UsableIf(actorTable.usableFunc):Use(actorTable.target, actorTable.condition)
    end
    if self:GetActor().type == "action" then
        ---@cast actorTable APLActorActionTable
        -- print("Bastion: APL:Execute: Executing action " .. actorTable.action)
        self:GetActor().cb()
    end
    if actorTable.type == "variable" then
        ---@cast actorTable APLActorVariableTable
        -- print("Bastion: APL:Execute: Setting variable " .. actorTable.variable)
        actorTable._apl.variables[actorTable.variable] = actorTable.cb(actorTable._apl)
    end
    return false
end

-- has traits
---@return boolean
function APLActor:HasTraits()
    return #self.traits > 0
end

-- tostring
---@return string
function APLActor:__tostring()
    return string.format("Bastion.__APLActor(%s)", self.name)
end

-- APL (Attack priority list) class
---@class APL
---@field apl APLActor[]
---@field variables table<string, any>
---@field name string
---@field last { successful: { name: string, time: number, index: number }, attempted: { name: string, time: number, index: number } }
local APL = {}
APL.__index = APL

-- Constructor
---@param name string
---@return APL
function APL:New(name)
    local self = setmetatable({}, APL)

    self.apl = {}
    self.variables = {}
    self.name = name
    self.last = {}

    return self
end

-- Add a variable to the APL
---@param name string
---@param value any
function APL:SetVariable(name, value)
    self.variables[name] = value
end

-- Get and evaluate a variable
---@param name string
---@return boolean
function APL:GetVariable(name)
    return self.variables[name]
end

---@class APLActorVariableTable : APLActorTableBase
---@field variable string
---@field cb fun(...):any
---@field _apl APL

-- Add variable
---@param name string
---@param cb fun(...):any
---@return APLActor
function APL:AddVariable(name, cb)
    local actor = APLActor:New({
        type = "variable",
        variable = name,
        cb = cb,
        _apl = self,
    })
    table.insert(self.apl, actor)
    return actor
end

---@class APLActorActionTable : APLActorTableBase
---@field action string
---@field cb fun(...):any

-- Add a manual action to the APL
---@param action string
---@param cb fun(...):any
---@return APLActor
function APL:AddAction(action, cb)
    local actor = APLActor:New({
        type = "action",
        action = action,
        cb = cb,
    })
    table.insert(self.apl, actor)
    return actor
end

---@class APLActorSpellTable : APLActorTableBase
---@field spell Spell
---@field condition? string|fun(self: Spell): boolean
---@field castableFunc false | fun(self: Spell): boolean
---@field target Unit | false
---@field onCastFunc false | fun(self: Spell):any

-- Add a spell to the APL
---@param spell Spell
---@param condition? string|fun(self: Spell):boolean
---@return APLActor
function APL:AddSpell(spell, condition)
    local castableFunc = spell.CastableIfFunc
    local onCastFunc = spell.OnCastFunc
    local target = spell:GetTarget()

    local actor = APLActor:New({
        type = "spell",
        spell = spell,
        condition = condition,
        castableFunc = castableFunc,
        target = target,
        onCastFunc = onCastFunc,
    })

    table.insert(self.apl, actor)

    return actor
end

---@class APLActorItemTable : APLActorTableBase
---@field item Item
---@field condition? string | fun(self: Item): boolean
---@field usableFunc false | fun(...): boolean
---@field target Unit | nil

-- Add an item to the APL
---@param item Item
---@param condition? string | fun(self: Item): boolean
---@return APLActor
function APL:AddItem(item, condition)
    local usableFunc = item.UsableIfFunc
    local target = item:GetTarget()

    local actor = APLActor:New({
        type = "item",
        item = item,
        condition = condition,
        usableFunc = usableFunc,
        target = target,
    })

    table.insert(self.apl, actor)

    return actor
end

---@class APLActorAPLTable : APLActorTableBase
---@field apl APL
---@field condition? fun(...):boolean

-- Add an APL to the APL (for sub APLs)
---@param apl APL
---@param condition fun(...):boolean
---@return APLActor
function APL:AddAPL(apl, condition)
    if not condition then
        error("Bastion: APL:AddAPL: No condition for APL " .. apl.name)
    end
    local actor = APLActor:New({
        type = "apl",
        apl = apl,
        condition = condition,
    })
    table.insert(self.apl, actor)
    return actor
end

-- Execute the APL
function APL:Execute()
    for i, actor in ipairs(self.apl) do
        self.last.attempted = {
            name = actor.name,
            time = GetTime(),
            index = i,
        }
        if (not actor:HasTraits() or actor:Evaluate()) and actor:Execute() then
            self.last.successful = {
                name = actor.name,
                time = GetTime(),
                index = i,
            }
            return true
        end
    end
end

---@class APLActorSequencerTable : APLActorTableBase
---@field sequencer Sequencer
---@field condition? fun(...):boolean

-- Add a Sequencer to the APL
---@param sequencer Sequencer
---@param condition fun(...):boolean
---@return APLActor
function APL:AddSequence(sequencer, condition)
    local actor = APLActor:New({
        type = "sequencer",
        sequencer = sequencer,
        condition = condition,
    })
    table.insert(self.apl, actor)
    return actor
end

-- tostring
---@return string
function APL:__tostring()
    return "Bastion.__APL(" .. self.name .. ")"
end

return APL, APLActor, APLTrait
